@@ -3,7 +3,7 @@ from __future__ import absolute_import
 
 import json
 import keyword
-import requests
+from requests import RequestException as requests_RequestException, ConnectionError as requests_ConnectionError, exceptions as requests_exceptions, codes as requests_codes
 
 from .network import CFnetwork
 from .logging_helper import CFlogger
@@ -12,6 +12,7 @@ from .read_configs import read_configs
 from .api_v4 import api_v4
 from .api_extras import api_extras
 from .api_decode_from_web import api_decode_from_web
+from .api_decode_from_openapi import api_decode_from_openapi
 from .exceptions import CloudFlareError, CloudFlareAPIError, CloudFlareInternalError
 
 BASE_URL = 'https://api.cloudflare.com/client/v4'
@@ -193,21 +194,29 @@ class CloudFlare():
                     if parts[4]:
                         url += '/' + parts[4]
 
+            if files and data:
+                # Can't send data and form data - so move data into files and send as multipart/form-data
+                new_files = []
+                new_files += [(f, (files[f].name, files[f])) for f in files]
+                new_files += [(d, (None, data[d])) for d in data]
+                files = tuple(new_files)
+                data = None
+
             if self.logger:
                 msg = build_curl(method, url, headers, params, data, files)
                 self.logger.debug('Call: emulated curl command ...\n%s', msg)
 
             try:
                 response = self.network(method, url, headers, params, data, files)
-            except requests.RequestException as e:
+            except requests_RequestException as e:
                 if self.logger:
                     self.logger.debug('Call: requests exception! "%s"', e)
                 raise CloudFlareAPIError(0, e)
-            except requests.ConnectionError as e:
+            except requests_ConnectionError as e:
                 if self.logger:
                     self.logger.debug('Call: requests connection exception! "%s"', e)
                 raise CloudFlareAPIError(0, 'connection error')
-            except requests.exceptions.Timeout as e:
+            except requests_exceptions.Timeout as e:
                 if self.logger:
                     self.logger.debug('Call: requests timeout exception! "%s"', e)
                 raise CloudFlareAPIError(0, 'connection timeout')
@@ -297,7 +306,7 @@ class CloudFlare():
                 except ValueError:
                     if response_data == '':
                         # This should really be 'null' but it isn't. Even then, it's wrong!
-                        if response_code == requests.codes.ok:
+                        if response_code == requests_codes.ok:
                             # 200 ok
                             response_data = {'success': True,
                                              'result': None}
@@ -320,7 +329,7 @@ class CloudFlare():
                                 self.logger.debug('Response data not JSON: %r', response_data)
                             raise CloudFlareAPIError(0, 'JSON parse failed - report to Cloudflare.')
 
-                if response_code == requests.codes.ok:
+                if response_code == requests_codes.ok:
                     # 200 ok - so nothing needs to be done
                     pass
                 else:
@@ -329,7 +338,7 @@ class CloudFlare():
                     pass
             elif response_type == 'application/octet-stream' and isinstance(response_data, (int, float)):
                 # It's binary data
-                if response_code == requests.codes.ok:
+                if response_code == requests_codes.ok:
                     # 200 ok
                     response_data = {'success': True,
                                      'result': response_data}
@@ -345,7 +354,7 @@ class CloudFlare():
                 try:
                     response_data = json.loads(response_data)
                     if not isinstance(response_data, (dict)) or 'success' not in response_data:
-                        if response_code == requests.codes.ok:
+                        if response_code == requests_codes.ok:
                             # 200 ok
                             response_data = {'success': True,
                                              'result': response_data}
@@ -357,7 +366,7 @@ class CloudFlare():
                 except ValueError:
                     # So it wasn't JSON - moving on as if it's text!
                     # A single value is returned (vs an array or object)
-                    if response_code == requests.codes.ok:
+                    if response_code == requests_codes.ok:
                         # 200 ok
                         response_data = {'success': True, 'result': response_data}
                     else:
@@ -377,7 +386,7 @@ class CloudFlare():
                 except ValueError:
                     # So it wasn't JSON - moving on as if it's text!
                     # A single value is returned (vs an array or object)
-                    if response_code == requests.codes.ok:
+                    if response_code == requests_codes.ok:
                         # 200 ok
                         response_data = {'success': True, 'result': response_data}
                     else:
@@ -387,7 +396,7 @@ class CloudFlare():
                                          'result': response_data}
             elif response_type in ['text/javascript', 'application/javascript']:
                 # used by Cloudflare workers
-                if response_code == requests.codes.ok:
+                if response_code == requests_codes.ok:
                     # 200 ok
                     response_data = {'success': True,
                                      'result': str(response_data)}
@@ -398,7 +407,7 @@ class CloudFlare():
                                      'result': str(response_data)}
             elif response_type == 'text/html':
                 # used by media for preview
-                if response_code == requests.codes.ok:
+                if response_code == requests_codes.ok:
                     # 200 ok
                     response_data = {'success': True,
                                      'result': str(response_data)}
@@ -411,7 +420,7 @@ class CloudFlare():
             else:
                 # Assuming nothing - but continuing anyway
                 # A single value is returned (vs an array or object)
-                if response_code == requests.codes.ok:
+                if response_code == requests_codes.ok:
                     # 200 ok
                     response_data = {'success': True,
                                      'result': str(response_data)}
@@ -535,7 +544,15 @@ class CloudFlare():
 
             # base url isn't enough; we need less
             url = '/'.join(self.base_url.split('/')[0:3])
+            return self._read_from_web(url)
 
+        def api_from_openapi(self, url):
+            """ Cloudflare v4 API"""
+
+            return self._read_from_web(url)
+
+        def _read_from_web(self, url):
+            """ Cloudflare v4 API"""
             try:
                 if self.logger:
                     self.logger.debug('Call: doit!')
@@ -894,6 +911,11 @@ class CloudFlare():
 
         return api_decode_from_web(self._base.api_from_web())
 
+    def api_from_openapi(self, url):
+        """ Cloudflare v4 API"""
+
+        return api_decode_from_openapi(self._base.api_from_openapi(url))
+
     def __init__(self, email=None, key=None, token=None, certtoken=None, debug=False, raw=False, use_sessions=True, profile=None, base_url=None):
         """ Cloudflare v4 API"""
 
